import warnings
import sys
from copy import copy

from functools import wraps, partial
from web3._utils.abi import filter_by_name
from web3.contract.contract import ContractFunction
from web3.types import ABI
from eth_typing import HexStr, AnyAddress

from chain_aide.base.module import Module
from chain_aide.utils import contract_call, contract_transaction


class Contract(Module):
    abi: ABI = None
    bytecode: HexStr = None
    address: AnyAddress = None

    def init(self,
             abi,
             bytecode=None,
             address=None,
             ):
        self.__build_contract(abi, bytecode, address)
        return copy(self)

    def deploy(self,
               abi,
               bytecode,
               txn=None,
               private_key=None,
               *init_args,
               **init_kwargs):
        if self.address:
            warnings.warn(f'contract {self.address} already exists, it will be replaced.', RuntimeWarning)

        _temp_origin = self.aide.web3.eth.contract(abi=abi, bytecode=bytecode)
        txn = _temp_origin.constructor(*init_args, **init_kwargs).build_transaction(txn)
        tx_hash = self.aide.send_transaction(txn, private_key)
        receipt = self.aide.eth.wait_for_transaction_receipt(tx_hash, timeout=20)

        address = receipt.get('contractAddress')
        if not address:
            raise Exception(f'deploy contract failed, because: {receipt}.')

        self.__build_contract(abi, bytecode, address)

        return copy(self)

    def __build_contract(self,
                         abi,
                         bytecode=None,
                         address=None,
                         ):
        self.abi = abi
        self.bytecode = bytecode
        self.address = address
        self._origin = self.aide.web3.eth.contract(address=self.address, abi=self.abi, bytecode=self.bytecode)
        self.functions = self._origin.functions
        self.events = self._origin.events
        self._set_functions(self._origin.functions)
        self._set_events(self._origin.events)
        self._set_fallback(self._origin.fallback)

    def _set_functions(self, functions):
        # 合约event和function不会重名，因此不用担心属性已存在
        for func in functions:
            warp_function = self._function_wrap(getattr(functions, func))
            setattr(self, func, warp_function)

    def _set_events(self, events):
        # 合约event和function不会重名，因此不用担心属性已存在
        for event in events:
            # 通过方法名获取方法
            warp_event = self._event_wrap(event)
            setattr(self, event.event_name, warp_event)

    def _set_fallback(self, fallback):
        if type(fallback) is ContractFunction:
            warp_fallback = self._fallback_wrap(fallback)
            setattr(self, fallback, warp_fallback)
        else:
            self.fallback = fallback

    def _function_wrap(self, func):
        fn_abis = filter_by_name(func.fn_name, self.abi)
        if len(fn_abis) == 0:
            raise ValueError('The method ABI is not found.')

        # 对于重载方法，仅取其一个，但需要其方法类型相同
        # todo: 此处理存在隐患
        fn_abi = fn_abis[0]
        for _abi in fn_abis:
            if _abi.get('stateMutability') != fn_abi.get('stateMutability'):
                raise ValueError('override method are of different types')

        # 忽略首个参数 'self'，以适配公共合约包装类
        def fit_func(__self__, *args, **kwargs):
            return func(*args, **kwargs)

        if fn_abi.get('stateMutability') in ['view', 'pure']:
            return partial(contract_call(fit_func), self)
        else:
            return partial(contract_transaction()(fit_func), self)

    @staticmethod
    def _event_wrap(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func().process_receipt(*args, **kwargs)

        return wrapper

    @staticmethod
    def _fallback_wrap(func):
        return contract_transaction()(func)



import functools
import os

from web3 import Web3, HTTPProvider, WebsocketProvider, IPCProvider
from web3._utils.threads import Timeout


def get_web3(uri, chain_id=None, hrp=None, timeout=10, modules=None) -> Web3:
    """ 通过rpc uri，获取web3对象
    """
    if uri.startswith('http'):
        provider = HTTPProvider
    elif uri.startswith('ws'):
        provider = WebsocketProvider
    elif uri.startswith('ipc'):
        provider = IPCProvider
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    with Timeout(timeout) as t:
        while True:
            web3 = Web3(provider(uri), modules=modules)
            if web3.is_connected():
                break
            t.sleep(1)

    return web3


def contract_call(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return func(self, *args, **kwargs).call()

    return wrapper


# 合约交易装饰器，仅用于接受参数
def contract_transaction(default_txn=None):
    # 实际装饰器
    def decorator(func):

        @functools.wraps(func)
        def wrapper(self, *args, txn: dict = None, private_key=None, **kwargs):
            """
            """
            # 合并交易体
            if txn:
                if default_txn:
                    default_txn.update(txn)
                    txn = default_txn
            else:
                txn = default_txn if default_txn else {}

            # 填充from地址，以免合约交易在预估gas时检验地址失败
            if not txn.get('from'):
                account = self.aide.eth.account.from_key(private_key,
                                                            hrp=self.aide.hrp) if private_key else self.aide.default_account
                if account:
                    txn['from'] = account.address

            # 构造合约方法对象
            if func.__name__ == 'fit_func':
                # solidity合约方法不传入private key参数，避免abi解析问题
                fn = func(self, *args, **kwargs)
            else:
                # 内置合约有时候需要用到私钥信息，用于生成参数的默认值，如：staking
                fn = func(self, *args, private_key=private_key, **kwargs)

            # 构建合约交易体dict
            txn = fn.build_transaction(txn)

            return self._transaction_handler_(txn, private_key=private_key)

        return wrapper

    return decorator


def execute_cmd(cmd):
    r = os.popen(cmd)
    out = r.read()
    r.close()
    return out



from chain_aide.base.module import Module


class Transfer(Module):
    transferGas: int = 21000

    def transfer(self, to_address, amount, txn=None, private_key=None):
        """ 发送转账交易
        """
        base_txn = {
            "to": to_address,
            "gasPrice": self.aide.eth.gas_price,
            "gas": self.transferGas,
            "data": '',
            "chainId": self.aide.chain_id,
            "value": amount,
        }

        if txn:
            base_txn.update(txn)

        return self._transaction_handler_(base_txn, private_key=private_key)

    def get_balance(self, address, block_identifier=None):
        """ 查询自由金额的余额
        """
        return self.aide.eth.get_balance(address, block_identifier)



from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from chain_aide import Aide


class Module:

    def __init__(self, aide: "Aide", module_type: Literal['', 'inner-contract'] = ''):
        self.aide = aide
        # 模块类型，当前仅用于判断是否自动解析内置合约的事件
        self.module_type = module_type

    def _transaction_handler_(self,
                              txn,
                              private_key=None,
                              ):
        """
        发送签名交易，并根据结果类型获取交易结果

        Args:
            txn: 要发送的交易dict
            private_key: 用于签名交易的私钥
        """
        if self.aide.result_type == "txn":
            return txn

        tx_hash = self.aide.send_transaction(txn, private_key=private_key)

        if self.aide.result_type == 'hash':
            return tx_hash

        receipt = self.aide.web3.eth.wait_for_transaction_receipt(tx_hash)
        if type(receipt) is bytes:
            receipt = receipt.decode('utf-8')

        if self.aide.result_type == 'receipt' or self.aide.result_type == 'auto':
            return receipt

        raise ValueError(f'unsupported module type {self.module_type} or unknown result type {self.aide.result_type}')



import time
from typing import Literal

from loguru import logger
from web3.main import get_default_modules
from web3.middleware import geth_poa_middleware
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_utils import to_checksum_address, combomethod
from chain_aide.transfer import Transfer
from chain_aide.contract import Contract
from chain_aide.utils import get_web3


def get_modules(exclude: list = None):
    """ 排除节点关闭的API
    """
    if not exclude:
        exclude = []

    modules = get_default_modules()
    if 'admin' in exclude:
        modules['node'][1].pop('admin')
    if 'debug' in exclude:
        modules.pop('debug')

    return modules


class Aide:
    """ 主类，功能如下：
    1. 各个子模块的集合体，通过它来调用子模块发送交易
    2. 持有设置数据，如：默认交易账户、经济模型数据、返回结果类型等
    3. 包含一些常用方法，如：创建账户、等待块高/周期、区块解码等
    """

    def __init__(self, uri: str):
        """
        Args:
            uri: 节点开放的RPC链接
        """
        self.uri = uri
        self.default_account: LocalAccount = None  # 发送签名交易时适用的默认地址
        self.result_type = 'auto'  # 交易返回的结果类型，包括：auto, txn, hash, receipt
        # web3相关设置
        self.web3 = get_web3(uri)
        self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.chain_id = self.web3.eth.chain_id
        self.account = Account()
        # 设置模块
        self.__init_web3__()
        self.__init_module__()

    def __init_web3__(self):
        """ 设置web相关模块
        """
        self.eth = self.web3.eth
        self.txpool = self.web3.geth.txpool
        self.personal = self.web3.geth.personal
        self.admin = self.web3.geth.admin

    def __init_module__(self):
        self.transfer = Transfer(self)
        self.contract = Contract(self)

    @combomethod
    def create_account(self, is_hd=False):
        """ 创建账户
        """
        self.account._use_unaudited_hdwallet_features = is_hd
        return Account.create()

    @combomethod
    def create_keystore(self, passphrase, key=None):
        """ 创建钱包文件
        """
        pass

    def set_default_account(self, account: LocalAccount):
        """ 设置发送交易的默认账户
        """
        self.default_account = account

    def set_result_type(self,
                        result_type: Literal['auto', 'txn', 'hash', 'receipt']
                        ):
        """ 设置返回结果类型，建议设置为auto
        """
        self.result_type = result_type

    def send_transaction(self, txn: dict, private_key=None):
        """ 签名交易并发送，返回交易hash
        """
        if not private_key and self.default_account:
            private_key = self.default_account.key

        if not txn.get('nonce'):
            account = self.eth.account.from_key(private_key)
            txn['nonce'] = self.eth.get_transaction_count(account.address)

        signed_txn = self.eth.account.sign_transaction(txn, private_key)
        tx_hash = self.eth.send_raw_transaction(signed_txn.rawTransaction)
        return bytes(tx_hash).hex()

    def wait_block(self, to_block, time_out=None):
        """ 等待块高
        """
        current_block = self.eth.block_number
        time_out = time_out or (to_block - current_block) * 3

        for i in range(time_out):
            time.sleep(1)
            current_block = self.eth.block_number

            if i % 10 == 0:
                logger.info(f'waiting block: {current_block} -> {to_block}')

            # 等待确定落链
            if current_block > to_block:
                logger.info(f'waiting block: {current_block} -> {to_block}')
                return

        raise TimeoutError('wait block timeout!')

    @combomethod
    def to_checksum_address(self, address):
        """ 任意地址转换为checksum地址
        """
        return to_checksum_address(address)

    @combomethod
    def to_base58_address(self, address):
        """ 任意地址转换为形式的base58地址
        注意：非标准base58地址
        """
        pass

    # def ec_recover(self, block_identifier):
    #     """ 使用keccak方式，解出区块的签名节点公钥
    #     """
    #     block = self.web3.eth.get_block(block_identifier)
    #
    #     extra = block.proofOfAuthorityData[:32]
    #     sign = block.proofOfAuthorityData[32:]
    #     raw_data = [bytes.fromhex(remove_0x_prefix(block.parentHash.hex())),
    #                 to_canonical_address(block.miner),
    #                 bytes.fromhex(remove_0x_prefix(block.stateRoot.hex())),
    #                 bytes.fromhex(remove_0x_prefix(block.transactionsRoot.hex())),
    #                 bytes.fromhex(remove_0x_prefix(block.receiptsRoot.hex())),
    #                 bytes.fromhex(remove_0x_prefix(block.logsBloom.hex())),
    #                 block.number,
    #                 block.gasLimit,
    #                 block.gasUsed,
    #                 block.timestamp,
    #                 extra,
    #                 bytes.fromhex(remove_0x_prefix(block.nonce.hex()))
    #                 ]
    #     hash_bytes = HexBytes(keccak(rlp.encode(raw_data)))
    #     signature_bytes = HexBytes(sign)
    #     signature_bytes_standard = to_standard_signature_bytes(signature_bytes)
    #     signature = Signature(signature_bytes=signature_bytes_standard)
    #     return remove_0x_prefix(HexStr(signature.recover_public_key_from_msg_hash(hash_bytes).to_hex()))
